
AVCodec do(
	newSlot("codecName", "mp2")
	newSlot("inputCoro")
	newSlot("internalCoro")
	newSlot("streamDestination", AudioDevice)
	
	willProcess := method(
		//writeln("AVCodec willProcess")
		if(inputBuffer isEmpty, 
			//writeln("AVCodec inputBuffer empty - pausing internalCoro")
			setInternalCoro(Coroutine currentCoroutine)
			internalCoro pause
		)
	)

	didProcess := method(
		//writeln("AVCodec didProcess")
		if(streamDestination, streamDestination write(outputBuffer))
		outputBuffer empty
		inputCoro ?resumeLater
	)
	
	write := method(data,
		inputBuffer appendSeq(data)
		setInputCoro(Coroutine currentCoroutine)
		//writeln("AVCodec write - resuming internalCoro")
		internalCoro resume // this will pause the current coro
	)
)

AudioDevice do(
	write := method(data,
		asyncWrite(data)
		//writeln("AudioDevice write yielding until data is needed")
		while(needsData not, yield)
		//writeln("AudioDevice data needed")
	)
)

SampleRateConverter do(
	newSlot("streamDestination")
	
	write := method(data,
		setOutputToInputRatio(44100 / call sender sampleRate) 
		inputBuffer appendSeq(data)
		//writeln("SampleRateConverter process")
		process
		//writeln("SampleRateConverter process done")
		if(streamDestination, 
			streamDestination write(outputBuffer)
			outputBuffer empty
		)
	)
)


AudioMixer := Object clone do(
	init := method(
		self sources := List clone
		self mixed := Seq clone
	)
	
	newSlot("streamDestination", AudioDevice)
	newSlot("processedSamples", 0)
	newSlot("isRunning", false)
	
	appendSource := method(source, sources append(source))
	removeSource := method(source, sources remove(source))
	
	process := method(sampleCount,
		byteCount := sampleCount * 8
		mixed setSize(byteCount)
		mixed zero
		sources foreach(source, 
			while(source outputBuffer < byteCount, source read)
			mixed float32ArrayAdd(source outputBuffer)
			source outputBuffer removeSlice(0, byteCount)
		)
		mixed float32ArrayMultiplyByScalar(sources size)
		streamDestination write(mixed)
		processedSamples = processedSamples + sampleCount
	)

	start := method(
		setIsRunning(true)
		while(isRunning, process(22050)) // 1/16th of a second
	)
	
	stop := method(
		setIsRunning(false)
	)
)

LibSndFile do(
	newSlot("path")
	newSlot("sampleRate")
	newSlot("channels")
	newSlot("format")
)

Sound := Object clone do(
	newSlot("path")
	newSlot("buffer")
	newSlot("sampleRate")
	newSlot("channels")
	newSlot("format")
	load := method(
		lsf := LibSndFile clone
		setBuffer(lsf outputBuffer)
		lsf setPath(path)
		lsf openForReading
		while(lsf read(16384), nil)
		setSampleRate(lsf sampleRate)
		setChannels(lsf channels)
		lsf close
		self
	)
)
