/*  copyright: Steve Dekorte, 2002
 *  All rights reserved. See _BSDLicense.txt.
 *
 *  An object for loading various image formats using
 *  image libraries. As well as manipulating and displaying 
 *  the decoded image data using OpenGL.
 *
 *  It can use GL to scale, composite and convert the image format.
 */
 
#ifndef GLIMAGE_DEFINED
#define GLIMAGE_DEFINED 1

#include <Common.h>
#include <ByteArray.h>

#if defined(__APPLE__) || defined(MACOSX)
#include <GLUT/glut.h>
#include <OpenGL/glu.h>
#include <OpenGL/gl.h>
/*
#include <OpenGL/OpenGL.h>
#include <OpenGL/glext.h>
*/
#else
#ifdef _WIN32
#include <windows.h>
#endif
#include <GL/glu.h>
#include <GL/gl.h>
#endif


typedef struct
{
  char *path;
  char *fileType;
  ByteArray *byteArray;
  unsigned char ownsByteArray;
  int width;
  int height;
  int format;
  char *error;
  int textureId;
  int textureWidth;
  int textureHeight;
  
  int originalWidth;
  int originalHeight;
  
  float encodingQuality;
  int decodingWidthHint;
  int decodingHeightHint;
} GLImage;

GLImage *GLImage_new(void);
GLImage *GLImage_newWithPath_(char *fname);
void GLImage_free(GLImage *self);

void GLImage_setData_width_height_componentCount_(GLImage *self, ByteArray *ba, int width, int height, int componentCount);

ByteArray *GLImage_byteArray(GLImage *self);
void GLImage_setExternalByteArray_(GLImage *self, ByteArray *ba);

void GLImage_path_(GLImage *self, const char *path);
char *GLImage_path(GLImage *self);

void GLImage_fileType_(GLImage *self, const char *fileType);
char *GLImage_fileType(GLImage *self);

void GLImage_error_(GLImage *self, const char *path);
char *GLImage_error(GLImage *self);

void GLImage_load(GLImage *self);
void GLImage_save(GLImage *self);

int GLImage_width(GLImage *self);
void GLImage_width_(GLImage *self, int w);
int GLImage_height(GLImage *self);
void GLImage_height_(GLImage *self, int h);

int GLImage_format(GLImage *self);
void GLImage_format_(GLImage *self, int f);
int GLImage_componentCountForFormat_(GLImage *self, int format);
int GLImage_formatForComponentCount_(GLImage *self, int componentCount) ;
int GLImage_componentCount(GLImage *self);

int GLImage_sizeInBytes(GLImage *self);
unsigned char *GLImage_data(GLImage *self);
void GLImage_data_length_(GLImage *self, unsigned char *data, size_t length); 

// drawing 

void GLImage_draw(GLImage *self);

// manipulation

void GLImage_scaleTo(GLImage *self, int w, int h);
void GLImage_resizeTo(GLImage *self, int w, int h);
void GLImage_flipY(GLImage *self);
void GLImage_crop(GLImage *self, int x, int y, int w, int h);
void GLImage_composite_to_(GLImage *self, GLImage *other, int x, int y);
void GLImage_grab(GLImage *self, int x, int y, int w, int h, int format);

// textures 

GLenum GLImage_textureFormat(GLImage *self);
void GLImage_grabTexture(GLImage *self, int x, int y, int w, int h);
int GLImage_textureId(GLImage *self);
void GLImage_bindTexture(GLImage *self);
void GLImage_updateTexture(GLImage *self);
void GLImage_closeTexture(GLImage *self);
void GLImage_drawTexture(GLImage *self);
void GLImage_drawTextureArea(GLImage *self, int w, int h);

void GLImage_setParameters(GLImage *self);

void GLImage_resizeToPowerOf2(GLImage *self);

int GLImage_textureWidth(GLImage *self);
int GLImage_textureHeight(GLImage *self);

int GLImage_originalWidth(GLImage *self);
int GLImage_originalHeight(GLImage *self);

//  extras 

void GLImage_encodingQuality_(GLImage *self, float q);
float GLImage_encodingQuality(GLImage *self);

void GLImage_decodingWidthHint_(GLImage *self, int v);
int GLImage_decodingWidthHint(GLImage *self);

void GLImage_decodingHeightHint_(GLImage *self, int v);
int GLImage_decodingHeightHint(GLImage *self);

#endif

