/*  copyright: Steve Dekorte, 2002
 *  All rights reserved. See _BSDLicense.txt.
 *
 *  This is an object wrapper for libjpeg. 
 *
 *  It's possible to use an external ByteArray for loading by calling
 *  JPGImage_setExternalByteArray_(), but be carefull not to use the 
 *  JPGImage instance after freeing it's external byte array.
 */
 
#ifndef JPGIMAGE_DEFINED
#define JPGIMAGE_DEFINED 1

#include <ByteArray.h>
#include <jpeglib.h>

typedef struct
{
  char *path;
  int width;
  int height;
  int components;
  ByteArray *byteArray;
  unsigned char ownsByteArray;
  char *error;
  float quality; /* 0.0 - 1.0 */
  /*
  struct jpeg_decompress_struct cinfo;
  struct jpeg_error_mgr jerr;
  */
  int decodingWidthHint;
  int decodingHeightHint;
} JPGImage;

JPGImage *JPGImage_new(void);
JPGImage *JPGImage_newWithPath_(char *fname);
void JPGImage_free(JPGImage *self);

void JPGImage_path_(JPGImage *self, const char *path);
char *JPGImage_path(JPGImage *self);

void JPGImage_quality_(JPGImage *self, float q);
float JPGImage_quality(JPGImage *self);

void JPGImage_decodingWidthHint_(JPGImage *self, int w);
int JPGImage_decodingWidthHint(JPGImage *self);

void JPGImage_decodingHeightHint_(JPGImage *self, int h);
int JPGImage_decodingHeightHint(JPGImage *self);

void JPGImage_error_(JPGImage *self, const char *path);
char *JPGImage_error(JPGImage *self);

void JPGImage_load(JPGImage *self);
void JPGImage_save(JPGImage *self);

int JPGImage_width(JPGImage *self);
int JPGImage_height(JPGImage *self);

void JPGImage_width_(JPGImage *self, int w);
void JPGImage_height_(JPGImage *self, int h);
void JPGImage_components_(JPGImage *self, int c);

unsigned char JPGImage_isRGB8(JPGImage *self);
unsigned char JPGImage_isRGBA8(JPGImage *self);
unsigned char JPGImage_isL8(JPGImage *self);
unsigned char JPGImage_isLA8(JPGImage *self);
int JPGImage_components(JPGImage *self);

ByteArray *JPGImage_byteArray(JPGImage *self);

/* Whomever calls JPGImage_setExternalByteArray_ is responsible 
 * for freeing "ba" and for making sure that this
 * JPGImage is not still using it after it is freed.
 */
void JPGImage_setExternalByteArray_(JPGImage *self, ByteArray *ba); 

#endif 
