/*  copyright: Steve Dekorte, 2002
 *  All rights reserved. See _BSDLicense.txt.
 *
 *  This is an object wrapper for libtiff. 
 *
 *  It's possible to use an external ByteArray for loading by calling
 *  TIFFImage_setExternalByteArray_(), but be carefull not to use the 
 *  TIFFImage instance after freeing it's external byte array.
 */
 
#ifndef TIFFIMAGE_DEFINED
#define TIFFIMAGE_DEFINED 1

#include <ByteArray.h>

typedef struct
{
  char *path;
  int width;
  int height;
  int components;
  ByteArray *byteArray;
  unsigned char ownsBuffer;
  char *error;
} TIFFImage;

TIFFImage *TIFFImage_new(void);
TIFFImage *TIFFImage_newWithPath_(char *fname);
void TIFFImage_free(TIFFImage *self);

void TIFFImage_path_(TIFFImage *self, const char *path);
char *TIFFImage_path(TIFFImage *self);

void TIFFImage_error_(TIFFImage *self, const char *path);
char *TIFFImage_error(TIFFImage *self);

void TIFFImage_load(TIFFImage *self);
void TIFFImage_save(TIFFImage *self);

int TIFFImage_width(TIFFImage *self);
int TIFFImage_height(TIFFImage *self);

void TIFFImage_width_(TIFFImage *self, int w);
void TIFFImage_height_(TIFFImage *self, int h);
void TIFFImage_components_(TIFFImage *self, int c);

unsigned char TIFFImage_isRGB8(TIFFImage *self);
unsigned char TIFFImage_isRGBA8(TIFFImage *self);
unsigned char TIFFImage_isL8(TIFFImage *self);
unsigned char TIFFImage_isLA8(TIFFImage *self);
int TIFFImage_components(TIFFImage *self);

ByteArray *TIFFImage_byteArray(TIFFImage *self);

/* Whomever calls TIFFImage_setExternalByteArray_ is responsible 
 * for freeing "ba" and for making sure that this
 * TIFFImage is not still using it after it is freed.
 */
void TIFFImage_setExternalByteArray_(TIFFImage *self, ByteArray *ba); 

#endif 
