/*#io
SkipDBM ioDoc(
           docCopyright("Steve Dekorte", 2002)
           docLicense("BSD revised")
		 docCategory("Databases")
           docDescription("SkipDB is a skip-list based key-value database. SkipDBM manages any number of skipdbs within the same file. The root skipdb can be accessed using the root method.")
docObject           */

#include "IoSkipDBM.h"
#include "IoSkipDB.h"
#include "IoObject.h"
#include "IoState.h"
#include "IoSeq.h"
#include "IoState.h"
#include "IoNumber.h"

#define SKIPDBM(self) ((SkipDBM *)(IoObject_dataPointer(self)))

IoTag *IoSkipDBM_tag(void *state)
{
	IoTag *tag = IoTag_newWithName_("SkipDBM");
	tag->state = state;
	tag->freeFunc    = (TagFreeFunc *)IoSkipDBM_free;
	tag->cloneFunc   = (TagCloneFunc *)IoSkipDBM_rawClone;
	//tag->markFunc    = (TagMarkFunc *)IoSkipDBM_mark;
	//tag->writeToStoreOnStreamFunc  = (TagWriteToStoreOnStreamFunc *)IoSkipDBM_writeToStore_stream_;
	///tag->readFromStoreOnStreamFunc = (TagReadFromStoreOnStreamFunc *)IoSkipDBM_readFromStore_stream_;
	return tag;
}

void IoSkipDBM_writeToStore_stream_(IoObject *self, IoStore *store, BStream *stream)
{
}

void IoSkipDBM_readFromStore_stream_(IoObject *self, IoStore *store, BStream *stream)
{
}

IoSkipDBM *IoSkipDBM_proto(void *state)
{
	IoMethodTable methodTable[] = {
	{"setPath",  IoSkipDBM_setPath},     
	{"path",  IoSkipDBM_path},     
	{"open",     IoSkipDBM_open},  
	{"close",    IoSkipDBM_close}, 
	{"isOpen",   IoSkipDBM_isOpen},     
	{"delete",   IoSkipDBM_delete},     
	{"root",     IoSkipDBM_root}, 
	{"begin",     IoSkipDBM_beginTransaction}, 
	{"commit",     IoSkipDBM_commitnTransaction}, 
	//{"at",       IoSkipDBM_at},    
	{"compact",  IoSkipDBM_compact},  
	{NULL, NULL},
	};
	
	IoObject *self = IoObject_new(state);
	self->tag = IoSkipDBM_tag(state);
	
	IoObject_setDataPointer_(self, SkipDBM_new());
	IoState_registerProtoWithFunc_((IoState *)state, self, IoSkipDBM_proto);
	
	IoObject_addMethodTable_(self, methodTable);
	return self;
}

IoSkipDBM *IoSkipDBM_rawClone(IoSkipDBM *proto) 
{ 
	IoObject *self = IoObject_rawClonePrimitive(proto);
	self->tag = proto->tag;
	IoObject_setDataPointer_(self, SkipDBM_new());
	return self; 
}

IoSkipDBM *IoSkipDBM_new(void *state)
{
	IoObject *proto = IoState_protoWithInitFunction_((IoState *)state, IoSkipDBM_proto);
	return IOCLONE(proto);
}

void IoSkipDBM_free(IoSkipDBM *self) 
{	
	SkipDBM_free(SKIPDBM(self));
}

/*
void IoSkipDBM_mark(IoSkipDBM *self) 
{ 
}
*/

// -------------------------------------------------------- 

IoObject *IoSkipDBM_setPath(IoObject *self, IoObject *locals, IoMessage *m)
{
	/*#io
	docSlot("setPath(aString)", "Sets the path to the dbm folder. Returns self.")
	*/
	IoSeq *v = IoMessage_locals_seqArgAt_(m, locals, 0);
	
	SkipDBM_setPath_(SKIPDBM(self), CSTRING(v));
	return self;
}


IoObject *IoSkipDBM_path(IoObject *self, IoObject *locals, IoMessage *m)
{
	/*#io
	docSlot("path", "Returns the path to the dbm. Returns self.")
	*/

	return IOSYMBOL(SkipDBM_path(SKIPDBM(self)));
}

IoObject *IoSkipDBM_open(IoObject *self, IoObject *locals, IoMessage *m)
{
	/*#io
	docSlot("open", "Opens the dbm. Returns self.")
	*/
	
	SkipDBM_open(SKIPDBM(self));
	return self;
}

IoObject *IoSkipDBM_close(IoObject *self, IoObject *locals, IoMessage *m)
{
	/*#io
	docSlot("close", "Closes the dbm.")
	*/
	
	SkipDBM_close(SKIPDBM(self));
	return self;
}

IoObject *IoSkipDBM_isOpen(IoObject *self, IoObject *locals, IoMessage *m)
{
	/*#io
	docSlot("delete", "Deletes the db. Returns self.")
	*/
	
	return IOBOOL(self, SkipDBM_isOpen(SKIPDBM(self)));
}

IoObject *IoSkipDBM_delete(IoObject *self, IoObject *locals, IoMessage *m)
{
	/*#io
	docSlot("delete", "Deletes the db. Returns self.")
	*/
	
	SkipDBM_delete(SKIPDBM(self));
	return self;
}

IoObject *IoSkipDBM_root(IoObject *self, IoObject *locals, IoMessage *m)
{
	/*#io
	docSlot("root", "Returns the root SkipDB.")
	*/
	IOASSERT(SkipDBM_isOpen(SKIPDBM(self)), "skipdbm not open");

	{
	SkipDB *sdb = SkipDBM_rootSkipDB(SKIPDBM(self));
	
	return IoSkipDB_newWithSDB(IOSTATE, sdb);
	//return self;
	}
}

IoObject *IoSkipDBM_at(IoObject *self, IoObject *locals, IoMessage *m)
{
	/*#io
	docSlot("at(pidNumber)", "Returns the SkipDB at the specified persistent ID or nil if it is not found.")
	*/
	/*
	PID_TYPE pid = IoMessage_locals_intArgAt_(m, locals, 0);
	
	SkipDB *sdb = SkipDBM_skipdbAtPid_(SKIPDBM(self));
	return IoSkipDB_newWithSDB(IOSTATE, sdb);
	*/
	IOASSERT(SkipDBM_isOpen(SKIPDBM(self)), "skipdbm not open");
	return self;
}

IoObject *IoSkipDBM_compact(IoObject *self, IoObject *locals, IoMessage *m)
{
	/*#io
	docSlot("compact", "Compacts the database. Returns self.")
	*/
	
	IOASSERT(SkipDBM_isOpen(SKIPDBM(self)), "skipdbm not open");
	SkipDBM_compact(SKIPDBM(self));
	return self;
}

IoObject *IoSkipDBM_beginTransaction(IoObject *self, IoObject *locals, IoMessage *m)
{
	/*#io
	docSlot("beginTransaction", "Begin a transaction. Returns self.")
	*/
	
	IOASSERT(SkipDBM_isOpen(SKIPDBM(self)), "skipdbm not open");
	SkipDBM_beginTransaction(SKIPDBM(self));
	return self;
}

IoObject *IoSkipDBM_commitnTransaction(IoObject *self, IoObject *locals, IoMessage *m)
{
	/*#io
	docSlot("commitTransaction", "Commit a transaction. Returns self.")
	*/
	
	IOASSERT(SkipDBM_isOpen(SKIPDBM(self)), "skipdbm not open");
	SkipDBM_commitTransaction(SKIPDBM(self));
	return self;
}


