/*
 docCopyright("Steve Dekorte", 2002)
 docLicense("BSD revised")
 docDescription("See the release notes for a list of folks that helped with this code.")
 */

#ifndef PORTABLESOCKET_DEFINED
#define PORTABLESOCKET_DEFINED 1

#include <sys/types.h>

#ifndef WIN32
#include <sys/uio.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#define myerrno errno
#else
#include <winsock2.h>
#ifndef EISCONN
#define EISCONN WSAEISCONN
#define ENOTCONN WSAENOTCONN
#define myerrno WSAGetLastError()
#endif
#endif

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>

#ifdef sun
#include <sys/file.h> /* for FASYNC */
#endif

#ifndef SOMAXCONN
#define SOMAXCONN 128
#endif

#ifdef WIN32
typedef SOCKET SOCKET_DESCRIPTOR;
#else
typedef int SOCKET_DESCRIPTOR;
#endif

#include "IPAddress.h"
#include "ByteArray.h"

typedef struct
{
	SOCKET_DESCRIPTOR fd;
	//char *error;
} Socket;

void Socket_GlobalInit(void);
void Socket_GlobalCleanup(void);
long Socket_SetDescriptorLimitToMax(void);

Socket *Socket_new(void);
void Socket_free(Socket *self);

//void Socket_setError_(Socket *self, const char *e);
//const char *Socket_error(Socket *self);

void Socket_setDescriptor_(Socket *self, SOCKET_DESCRIPTOR fd);
SOCKET_DESCRIPTOR Socket_descriptor(Socket *self);;

int Socket_streamOpen(Socket *self);
int Socket_udpOpen(Socket *self);
//int Socket_openFifo(Socket *self, const char *path);
//int Socket_openReadOnlyNonBlockingAtPath_(Socket *self, char *path);
int Socket_isOpen(Socket *self);

int RawDescriptor_isValid(int fd);

int Socket_isValid(Socket *self);
int Socket_close(Socket *self);

int Socket_makeReusable(Socket *self);
int Socket_makeAsync(Socket *self);

int Socket_connectTo(Socket *self, IPAddress *address);
Socket *Socket_accept(Socket *self, IPAddress *address);
int Socket_bind(Socket *self, IPAddress *address);
int Socket_listen(Socket *self);

ssize_t Socket_streamRead(Socket *self, ByteArray *buffer, size_t readSize);
ssize_t Socket_streamWrite(Socket *self, ByteArray *buffer, size_t start, size_t writeSize);

ssize_t Socket_udpRead(Socket *self, IPAddress *address, ByteArray *buffer, size_t readSize);
ssize_t Socket_udpWrite(Socket *self, IPAddress *address, ByteArray *buffer, size_t start, size_t writeSize);

int Socket_isStream(Socket *self);
//size_t Socket_sendfile(Socket *self, int fd);

#endif
