/*#io
Thunder ioDoc(
		docCopyright("Prunedtree", 2004)
		docLicense("BSD revised")
		docCategory("Compiler")
		docDescription("""
""")
		*/

#include "IoThunder.h"
#include "IoState.h"
#include "IoNumber.h"
#include "IoSeq.h"

#define DATA(self) ((IoThunderData *)(IoObject_dataPointer(self)))

#define INT_ARG_0 (IoMessage_locals_intArgAt_(m, locals, 0))
#define INT_ARG_1 (IoMessage_locals_intArgAt_(m, locals, 1))
#define INT_ARG_2 (IoMessage_locals_intArgAt_(m, locals, 2))
#define INT_ARG_3 (IoMessage_locals_intArgAt_(m, locals, 3))

#define FLOAT_ARG_0 (IoMessage_locals_floatArgAt_(m, locals, 0))
#define FLOAT_ARG_1 (IoMessage_locals_floatArgAt_(m, locals, 1))
#define FLOAT_ARG_2 (IoMessage_locals_floatArgAt_(m, locals, 2))
#define FLOAT_ARG_3 (IoMessage_locals_floatArgAt_(m, locals, 3))

#define DOUBLE_ARG_0 (IoMessage_locals_doubleArgAt_(m, locals, 0))
#define DOUBLE_ARG_1 (IoMessage_locals_doubleArgAt_(m, locals, 1))
#define DOUBLE_ARG_2 (IoMessage_locals_doubleArgAt_(m, locals, 2))
#define DOUBLE_ARG_3 (IoMessage_locals_doubleArgAt_(m, locals, 3))

IoTag *IoThunder_tag(void *state)
{
	IoTag *tag = IoTag_newWithName_("Thunder");
	tag->state = state;
	tag->freeFunc  = (TagFreeFunc *)IoThunder_free;
	tag->cloneFunc = (TagCloneFunc *)IoThunder_rawClone;
	return tag;
}

IoThunder *IoThunder_rawClone(IoThunder *proto) 
{ 
	IoObject *self = IoObject_rawClonePrimitive(proto);
	IoObject_setDataPointer_(self, calloc(1, sizeof(IoThunderData)));
	return self; 
}

IoThunder *IoThunder_new(void *state)
{
	IoObject *proto = IoState_protoWithInitFunction_(state, IoThunder_proto);
	return IOCLONE(proto);
}

void IoThunder_free(IoThunder *self) 
{ 
	free(DATA(self)); 
}

// ----------------------------------------------------------- *

void thunder_put_op(char** code_ptr, char* op, ...); // int a, int b, int c)

IoObject *IoThunder_op(IoThunder *self, IoObject *locals, IoMessage *m)
{
	char *op = CSTRING(IoMessage_locals_seqArgAt_(m, locals, 0));
	
	if (DATA(self)->ip == NULL)
	{
		DATA(self)->ip = DATA(self)->code_base;
	}
	
	switch (IoMessage_argCount(m) - 1)
	{
		case 0:
			thunder_put_op(&(DATA(self)->ip), op);
			break;
		case 1:
			thunder_put_op(&(DATA(self)->ip), op, INT_ARG_1);
			break;
		case 2:
			thunder_put_op(&(DATA(self)->ip), op, INT_ARG_1, INT_ARG_2);
			break;
		case 3:
			thunder_put_op(&(DATA(self)->ip), op, INT_ARG_1, INT_ARG_2, INT_ARG_3);
			break;
		default:
			printf("IoThunder_op: too many args\n");
	}
	
	return self;
}

IoObject *IoThunder_calli(IoThunder *self, IoObject *locals, IoMessage *m)
{
	void *f = DATA(self)->code_base;
	int r = 0;
	
	switch (IoMessage_argCount(m))
	{
		case 0:
			r = ((int (*)(void))f)();
			break;
		case 1:
			r = ((int (*)(int))f)(INT_ARG_0);
			break;
		case 2:
			r = ((int (*)(int, int))f)(INT_ARG_0, INT_ARG_1);
			break;
		case 3:
			r = ((int (*)(int, int, int))f)(INT_ARG_0, INT_ARG_1, INT_ARG_2);
			break;
		default:
			printf("IoThunder_calli: too many args\n");
	}
	
	return IONUMBER(r);
}

IoObject *IoThunder_functionPointer(IoThunder *self, IoObject *locals, IoMessage *m)
{
	return IONUMBER((int)DATA(self)->code_base);
}

IoObject *IoThunder_instructionPointer(IoThunder *self, IoObject *locals, IoMessage *m)
{
	return IONUMBER((int)DATA(self)->ip);
}


// -------------------------------------------------------

IoThunder *IoThunder_proto(void *state)
{
	IoThunder *self = IoObject_new(state);
	self->tag = IoThunder_tag(state);
	
	IoObject_setDataPointer_(self, calloc(1, sizeof(IoThunderData)));
	
	IoState_registerProtoWithFunc_(state, self, IoThunder_proto);

	{
		IoMethodTable methodTable[] = {
		{"functionPointer", IoThunder_functionPointer},
		{"instructionPointer", IoThunder_instructionPointer},
		{"op", IoThunder_op},
		{"calli", IoThunder_calli},
		{NULL, NULL}
		};
		IoObject_addMethodTable_(self, methodTable);
	}
	
	return self;
}


// thunder.c - thunder, portable JIT assembler - rev2
// insert rev. BSD here

#include <stdio.h>
#include <string.h>
#include <stdarg.h>

typedef struct thunder_encoding thunder_encoding;

struct thunder_encoding
{
	char* mnemonic;
	char* code;
	char flag;
};

thunder_encoding list[] =
{
	/*	operands :
		      =>		(none)
			r => reg32
			i => imm32
			p => [reg32]
			m => [mem32]
			a => relative address
			b => imm8

		ModR/M flag :
			  => none
			+ => none, add register to last byte
		 0..9 => spare=0..9
			r => spare=register

	*/

	{	"add_rr","\x03",	'r'	},	// a = a + b
	{	"add_ri","\x81",	'0'	},	// a = a + b
	{	"sub_rr","\x2B",	'r'	},	// a = a - b
	{	"sub_ri","\x81",	'5'	},	// a = a - b
	{	"mul_rr","\x0F\xAF",'r'	},	// a = a * b
	{	"and_rr","\x23",	'r'	},	// a = a & b
	{	"and_ri","\x81",	'4'	},	// a = a & b
	{	"or_rr", "\x0B",	'r'	},	// a = a | b
	{	"or_ri", "\x81",	'1'	},	// a = a | b
	{	"xor_rr","\x33",	'r'	},	// a = a ^ b
	{	"xor_ri","\x81",	'6'	},	// a = a ^ b
	{	"neg_r", "\xF7",	'3'	},	// a = - a
	{	"not_r", "\xF7",	'2'	},	// a = ~ a

	{	"push_r","\x50",	'+'	},	// push (reg) a
	{	"push_i","\x68",	' '	},	// push (imm) a
	{	"push_p","\xFF",	'6'	},	// push *(reg) a
	{	"push_m","\xFF",	'6'	},	// push *(imm) a
	{	"pop_r", "\x58",	'+'	},	// pop (reg) a

	{	"mov_rr", "\x8B",	'r'	},	// a = (reg) b
	{	"mov_ri", "\xB8",	'+'	},	// a = (imm) b
	{	"mov_rp", "\x8B",	'r'	},	// a <= *(reg) b
	{	"mov_pr", "\x89",	'r'	},	// a => *(reg) b
	{	"mov_rm", "\x8B",	'r'	},	// a <= *(imm) b
	{	"mov_mr", "\x89",	'r'	},	// a => *(imm) b
	{	"mov_rpi","\x8B",	'r'	},	// a <= *( (reg) b + (imm) c )
	{	"mov_pri","\x89",	'r'	},	// a => *( (reg) b + (imm) c )

	{	"cmp_rr","\x3B",	'r'	},	// flags = a - (reg) b
	{	"cmp_ri","\x81",	'7'	},	// flags = a - (imm) b

	{	"jlt_a", "\x0F\x8C",' '	},	// if ( <  ) goto ip+a
	{	"jle_a", "\x0F\x8E",' '	},	// if ( <= ) goto ip+a
	{	"jgt_a", "\x0F\x8F",' '	},	// if ( >  ) goto ip+a
	{	"jge_a", "\x0F\x8D",' '	},	// if ( >= ) goto ip+a
	{	"jeq_a", "\x0F\x84",' '	},	// if ( == ) goto ip+a
	{	"jne_a", "\x0F\x85",' '	},	// if ( != ) goto ip+a

	{	"call_a","\xE8",	' '	},	// function call to address a
	{	"call_r","\xFF",	'2'	},	// function call to register a
	{	"call_m","\xFF",	'2'	},	// function call to pointer a
	{	"jmp_a", "\xFF",	'4'	},	// unconditional jump to register a
	{	"jmp_r", "\xFF",	'4'	},	// unconditional jump to register a
	{	"ret",	 "\xC3",	' '	},	// return from subroutine

	{	"rdtsc",		"\x0F\x31",	'r'	},	

	// mmx/isse cpuset
	{	"emms",			"\x0F\x77",	' ' },
	{	"maskmovq_rr",	"\x0F\xF7",	'r' },
	{	"movd_rr",		"\x0F\x6E",	'r' },
	{	"movd_rm",		"\x0F\x6E",	'r' },
	{	"movd_mr",		"\x0F\x7E",	'r' },
	{	"movntq_rr",	"\x0F\xE7",	'r' },
	{	"movq_rr",		"\x0F\x6F",	'r' },
	{	"movq_rm",		"\x0F\x6F",	'r' },
	{	"movq_mr",		"\x0F\x7F",	'r' },
	{	"packssdw_rr",	"\x0F\x6B",	'r'	},	
	{	"packsswb_rr",	"\x0F\x63",	'r'	},	
	{	"packuswb_rr",	"\x0F\x67",	'r'	},	
	{	"paddb_rr",		"\x0F\xFC",	'r'	},
	{	"paddw_rr",		"\x0F\xFD",	'r'	},
	{	"paddd_rr",		"\x0F\xFE",	'r'	},
	{	"paddq_rr",		"\x0F\xD4",	'r' },
	{	"paddsb_rr",	"\x0F\xEC",	'r' },
	{	"paddsw_rr",	"\x0F\xED",	'r' },
	{	"paddusb_rr",	"\x0F\xDC",	'r' },
	{	"paddusw_rr",	"\x0F\xdD",	'r' },
	{	"pand_rr",		"\x0F\xDB",	'r' },
	{	"pandn_rr",		"\x0F\xDF",	'r' },
	{	"pavgb_rr",		"\x0F\xE0",	'r' },
	{	"pavgw_rr",		"\x0F\xE3",	'r' },
	{	"pavgusb_rr",	"\x0F\x0F",	'r' },
	{	"pcmpeqb_rr",	"\x0F\x74",	'r' },
	{	"pcmpeqw_rr",	"\x0F\x75",	'r' },
	{	"pcmpeqd_rr",	"\x0F\x76",	'r' },
	{	"pcmpgtb_rr",	"\x0F\x64",	'r' },
	{	"pcmpgtw_rr",	"\x0F\x65",	'r' },
	{	"pcmpgtd_rr",	"\x0F\x66",	'r' },
	{	"pextrw_rr",	"\x0F\xC5",	'r' },
	{	"pinsrw_rr",	"\x0F\xC4",	'r' },
	{	"pmaddwd_rr",	"\x0F\xF5",	'r' },
	{	"pmaxsw_rr",	"\x0F\xEE",	'r' },
	{	"pmaxub_rr",	"\x0F\xDE",	'r' },
	{	"pminsw_rr",	"\x0F\xEA",	'r' },
	{	"pminub_rr",	"\x0F\xDA",	'r' },
	{	"pmovmskb_rr",	"\x0F\xD7",	'r' },
	{	"pmulhuw_rr",	"\x0F\xE4",	'r' },
	{	"pmulhw_rr",	"\x0F\xE5",	'r' },
	{	"pmullw_rr",	"\x0F\xD5",	'r' },
	{	"por_rr",		"\x0F\xEB",	'r' },
	{	"prefetchnta_m","\x0F\x18",	'0' },
	{	"prefetcht0_m",	"\x0F\x18",	'1' },
	{	"prefetcht1_m",	"\x0F\x18",	'2' },
	{	"prefetcht2_m",	"\x0F\x18",	'3' },
	{	"psadbw_rr",	"\x0F\xF6",	'r' },
	{	"pshufw_rr",	"\x0F\x70",	'r' },
	{	"psllw_rr",		"\x0F\xF1",	'r' },
	{	"psllw_rb",		"\x0F\x71",	'6' },
	{	"pslld_rr",		"\x0F\xF2",	'r' },
	{	"pslld_rb",		"\x0F\x72",	'6' },
	{	"psllq_rr",		"\x0F\xF3",	'r' },
	{	"psllq_rb",		"\x0F\x73",	'6' },
	{	"psraw_rr",		"\x0F\xE1",	'r' },
	{	"psraw_rb",		"\x0F\x71",	'4' },
	{	"psrad_rr",		"\x0F\xE2",	'r' },
	{	"psrad_rb",		"\x0F\x72",	'4' },
	{	"psrlw_rr",		"\x0F\xD1",	'r' },
	{	"psrlw_rb",		"\x0F\x71",	'2' },
	{	"psrld_rr",		"\x0F\xD2",	'r' },
	{	"psrld_rb",		"\x0F\x72",	'2' },
	{	"psrlq_rr",		"\x0F\xD3",	'r' },
	{	"psrlq_rb",		"\x0F\x73",	'2' },
	{	"psubb_rr",		"\x0F\xF8",	'r' },
	{	"psubw_rr",		"\x0F\xF9",	'r' },
	{	"psubd_rr",		"\x0F\xFA",	'r' },
	{	"psubsb_rr",	"\x0F\xE8",	'r' },
	{	"psubsw_rr",	"\x0F\xE9",	'r' },
	{	"psubusb_rr",	"\x0F\xD8",	'r' },
	{	"psubusw_rr",	"\x0F\xD9",	'r' },
	{	"punpckhbw_rr",	"\x0F\x68",	'r' },
	{	"punpckhwd_rr",	"\x0F\x69",	'r' },
	{	"punpckhdq_rr",	"\x0F\x6A",	'r' },
	{	"punpcklbw_rr",	"\x0F\x60",	'r' },
	{	"punpcklwd_rr",	"\x0F\x61",	'r' },
	{	"punpckldq_rr",	"\x0F\x62",	'r' },
	{	"pxor_rr",		"\x0F\xEF",	'r' },

	{	NULL,	NULL,	0 } // end of table
};

void thunder_put_op(char** code_ptr, char* op, ...) // int a, int b, int c)
{
	va_list args;
	char* code = *code_ptr;
	int i,k,u;
	int argc;
	char argt[3];

	// find mnemonic
	for(i=0; list[i].mnemonic != 0; i++)
	{
		if (0==strcmp(list[i].mnemonic, op)) break;
	}

	if (0 == list[i].mnemonic) return;

	// get argc/argt (count & types)
	for(k=0; op[k] != 0; k++) 
	{
		if (op[k] == '_') break;
	}

	if (op[k] == 0) 
		argc = 0;
	else
		argc = strlen(op) - (k+1);

	for(u=0; u<argc; u++) argt[u] = op[k+1+u];

	if (argc>1 && argt[0] != 'r') // swap for store movs
	{
		char t = argt[0]; 
		argt[0] = argt[1];
		argt[1] = t;
	}
	
	{
		int mod=3,spare,rm;
		int flag = list[i].flag;
		int imm,put_imm=0;
		int relative=0;

		va_start( args, op );

		spare = list[i].flag-'0';

		for (u=0;u<argc;u++)
		{
			if ('r' == argt[u] || 'p' == argt[u])
			{
				if ('p' == argt[u]) mod = 0;
				if (u>0) spare = rm;
				rm = va_arg(args, int);
			}

			if ('i' == argt[u] || 'b' == argt[u] || 'm' == argt[u] || 'a' == argt[u])
			{
				if ('m' == argt[u]) { mod = 0; if (u>0) spare = rm; rm=5; }
				if ('a' == argt[u]) { relative=1; }
				if (u>1) mod = 2;
				put_imm = 4; // 32 bits
				if ('b' == argt[u]) put_imm = 1; // one byte
				imm = va_arg(args, int);
			}
		}

		// write opcode
		for(k=0;list[i].code[k];k++)
			*code++ = list[i].code[k];

		if ( flag == '+' )
			code[-1] += rm;

		// ModR/M byte
		if ( flag != ' ' && flag != '+' )
			*code++ = (mod<<6)+(spare<<3)+rm;

		// immediate
		if (put_imm)
		{
			if (relative) imm = imm - ((int)code + 4);
			if (put_imm==4)
				*(int*)code = imm;
			else 
				*code = imm;
			code += put_imm;
		}
	}

	*code_ptr = code;
	va_end( args );
	return;
}
