//List insertEven := method(arg, c := cursor; c insert(arg); c next; while(c next, c insert(arg); c next); self)
//List insertOdd  := method(arg, c := self cursor; while(c next, c insert(arg); c next); self)

Sequence prepend := method(s, s .. self)

Binding := Object clone do(
	cc := method(System getenv("CC") ifNilEval(return "cc"))
	cxx := method(System getenv("CXX") ifNilEval(return "g++"))
	ar := method(System getenv("AR") ifNilEval(return "ar"))
	ranlib := method(System getenv("RANLIB") ifNilEval(return "ranlib"))
	platform := System platform split at(0) asLowercase
	platformVersion := System platformVersion

	supportedOnPlatform := true

	frameworkSearchPaths := List clone
	frameworkSearchPaths append("/System/Library/Frameworks")
	frameworkSearchPaths append("/Library/Frameworks")
	frameworkSearchPaths append("/Local/Library/Frameworks")
//	frameworkSearchPaths append("~/Library/Frameworks")

	searchPrefixes := List clone
	searchPrefixes append(System installPrefix)
	searchPrefixes append("/usr")
	searchPrefixes append("/usr/X11R6")
	searchPrefixes append("/usr/local")
	searchPrefixes append("/usr/pkg")
	searchPrefixes append("/opt/local")
	searchPrefixes append("/sw")

	headerSearchPaths := List clone
	appendHeaderSearchPath := method(v, if(File clone setPath(v) exists, headerSearchPaths appendIfAbsent(v)))
	searchPrefixes foreach(i, searchPrefix, appendHeaderSearchPath(searchPrefix .. "/include"))

	libSearchPaths := List clone
	appendLibSearchPath := method(v, if(File clone setPath(v) exists, libSearchPaths appendIfAbsent(v)))
	searchPrefixes foreach(i, searchPrefix, appendLibSearchPath(searchPrefix .. "/lib"))

	init := method(
		self folder := Directory clone

		self depends := Object clone do(
			ports := List clone
//			dlls := List clone
			headers := List clone
			libs := List clone
			frameworks := List clone
			includes := List clone
			linkOptions := List clone
			addons := List clone
		)
	)

	pathForFramework := method(name,
		frameworkname := name .. ".framework"
		path := frameworkSearchPaths detect(path,
			Directory with(path .. "/" .. frameworkname) exists
		)
		path
	)

	pathForHeader := method(name,
		path := headerSearchPaths detect(path,
			File with(path .. "/" .. name) exists
		)
		path
	)

	pathForLib := method(name,
		libname := "lib" .. name
		path := libSearchPaths detect(path,
			File with(path .. "/" .. libname .. "." .. dllSuffix) exists or File with(path .. "/" .. libname .. ".a") exists
		)
		path
	)

	dependsOnBinding    := method(v, depends addons appendIfAbsent(v))
	dependsOnPort       := method(v, depends ports appendIfAbsent(v))
//	dependsOnDll        := method(v, depends dlls appendIfAbsent(v))
	dependsOnHeader     := method(v, depends headers appendIfAbsent(v))
	dependsOnLib        := method(v, depends libs appendIfAbsent(v))
	dependsOnFramework  := method(v, depends frameworks appendIfAbsent(v))
	dependsOnInclude    := method(v, depends includes appendIfAbsent(v))
	dependsOnLinkOption := method(v, depends linkOptions appendIfAbsent(v))

	dependsOnFrameworkOrLib := method(v, w,
		path := pathForFramework(v)
		if(path != nil, dependsOnFramework(v) ; appendHeaderSearchPath(path .. "/" .. v .. ".framework/Headers"), dependsOnLib(w))
	)

	optionallyDependsOnLib       := method(v, if(pathForFramework(v) != nil, dependsOnFramework(v)))
	optionallyDependsOnFramework := method(v, if(pathForLib(v) != nil, dependsOnLib(v)))

	missingFrameworks := method(
		depends frameworks select(p,
			if(pathForFramework(p) == nil,
				error := (self name .. " is missing " .. p .. " framework\n") print
				File clone openForAppending("errors") write(error) close
				self isAvailable := false
				true
			)
		)
	)
	missingHeaders := method(
		depends headers select(p,
			if(pathForHeader(p) == nil,
				error := (self name .. " is missing " .. p .. " header\n") print
				File clone openForAppending("errors") write(error) close
				self isAvailable := false
				true
			)
		)
	)
	missingLibs := method(
		depends libs select(p,
			if(pathForLib(p) == nil,
				error := (self name .. " is missing " .. p .. " library\n") print
				File clone openForAppending("errors") write(error) close
				self isAvailable := false
				true
			)
		)
	)
	hasDepends := method(missingFrameworks size == 0 and missingLibs size == 0 and missingHeaders size == 0)

	with := method(path,
		module := self clone
		module folder setPath(path)
		module
	)

	systemCall := method(s,
		if(trySystemCall(s) == 256, System exit(1))
	)

	trySystemCall := method(s,
		if(folder path != ".", s := "cd " .. folder path .. "; " .. s)
		writeln(s, "\n")
		result := System system("sh -c '" .. s .. "'")
		//if(result == 256, writeln("above error occured on command:\n", s, "\n"))
		result
	)

	// ------------------------------------

	name := method(folder name)
	oldDate := Date clone setYear(1970)
	
	libName := method("libIo" .. folder name ..  ".a")
//	dllName := method("libIo" .. folder name ..  "." .. dllSuffix)

	libFile := method(folder fileNamedOrNil(libName))
//	dllFile := method(folder fileNamedOrNil(dllName))
	objsFolder := method(self objsFolder := folder createSubdirectory("_build/objs"))
	sourceFolder := method(folder folderNamed("source"))
	cFiles := method(
		sourceFolder filesWithExtension("cpp") appendSeq(sourceFolder filesWithExtension("c")) appendSeq(sourceFolder filesWithExtension("m"))
	)

	newSlot("isServerBinding", false)

	libsFolder   := method(Directory with("libs"))
	addonsFolder := method(Directory with("addons"))

	includePaths := method(
		includePaths := List clone
		includePaths appendSeq(libsFolder folders map(path) map(p, Path with(p, "_build/headers")))
		includePaths appendSeq(depends addons map(n, "addons/" .. n .. "/_build/headers"))
		includePaths
	)

	build := method(options,
		//appendLibSearchPath(folder path)
		writeln("--- ", folder name, " ---")
		trySystemCall("mkdir -p _build/headers")
		trySystemCall("cp source/*.h _build/headers")

		generateInitFile

		options := options ifNilEval("")
		cFiles foreach(f,
			obj := f name replaceSeq(".cpp", ".o") replaceSeq(".c", ".o") replaceSeq(".m", ".o")
			objFile := objsFolder fileNamedOrNil(obj)
			if((objFile == nil) or(objFile lastDataChangeDate < f lastDataChangeDate)) then(

				includes := includePaths map(v, "-I" .. Path with("../../", v))
				includes appendSeq(headerSearchPaths map(v, "-I" .. v))

				s := cc .. " " .. options .. " " .. depends includes join(" ") .. " " .. includes join(" ") .. " -I. "
				if(list("cygwin", "windows") contains(platform) not, s = s .. "-fPIC ")
				s = s .. "-c -o _build/objs/" .. obj .. " source/" .. f name
				systemCall(s)
			)
		)

		buildLib
		buildDynLib
	)
	
	buildLib := method(
		trySystemCall("mkdir -p _build/lib")
		systemCall(ar .. " rcu _build/lib/" .. libName .. " _build/objs/*.o")
		systemCall(ranlib .. " _build/lib/" .. libName)
	)

	//dllSuffix := method(if(platform == "darwin", "bundle", "so"))
	dllSuffix := method(
		if(list("cygwin", "windows") contains(platform), return "dll")
		if(platform == "darwin", return "dylib")
		"so"
	)

	dllNameFor := method(s, "lib" .. s .. "." .. dllSuffix)

	//dllCommand := method(if(platform == "darwin", "-bundle", "-shared"))
	dllCommand := method(if(platform == "darwin", "-dynamiclib", "-shared"))

	buildDynLib := method(
		writeln("buildDynLib")
		trySystemCall("mkdir -p _build/dll")

		links := depends addons map(b, "-L../" .. b .. "/_build/dll")

		links appendSeq(depends addons map(v, "-lIo" .. v))
		if(platform != "darwin",
			links appendSeq(depends addons map(v, "-Wl,--rpath -Wl," .. System installPrefix .. "/lib/io/addons/" .. v .. "/_build/dll/"))
		)
		links appendSeq(libSearchPaths map(v, "-L" .. v))
		links appendSeq(depends libs map(v, "-l" .. v))
//		links appendSeq(depends dlls map(v, "-l" .. v))
//		links appendSeq(depends dlls map(v, "-I" .. v .. "/_build/headers"))
		links appendSeq(list("-L../../_build/dll", "-liovmall"))
		//links appendSeq(Directory with("libs") folders map(f, "-L../../" .. f path .. "/_build/dll -l" .. f name))

		links appendSeq(depends frameworks map(v, "-framework " .. v))

		libname := dllNameFor("Io" .. self name)

		s := ""
		if(platform == "darwin",
			links append("-flat_namespace")
			s := " -install_name " .. System installPrefix .. "/lib/io/addons/" .. self name .. "/_build/dll/" .. libname
		)

		systemCall(cc .. " " .. dllCommand .. " " .. s .. " -o _build/dll/" .. libname .. " _build/objs/*.o" .. " " .. links join(" "))
	)

	clean := method(
		writeln(folder name, " clean")
		trySystemCall("rm -rf _build")
		trySystemCall("rm -f source/Io*Init.c")
		self removeSlot("objsFolder")
	)

	ioCodeFolder := method(folder folderNamed("io"))
	ioFiles      := method(ioCodeFolder filesWithExtension("io"))
	initFileName := method("source/Io" .. folder name .. "Init.c")

	isStatic := false

	generateInitFile := method(
		if(folder folderNamed("source") filesWithExtension("m") size != 0, return)
		initFile := folder createFileNamed(initFileName)
		initFile remove open
		initFile write("#include \"IoState.h\"\n")
		initFile write("#include \"IoObject.h\"\n\n")

		sourceFiles := folder folderNamed("source") files
		iocFiles := sourceFiles select(f, f name beginsWithSeq("Io") and(f name endsWithSeq(".c")) and(f name containsSeq("Init") not) and(f name containsSeq("_") not))

		extraFiles := sourceFiles select(f, f name beginsWithSeq("Io") and(f name endsWithSeq(".c")) and(f name containsSeq("Init") not) and(f name containsSeq("_")))

		orderedFiles := List clone appendSeq(iocFiles)

		iocFiles foreach(f,
			//writeln("checking for depends on ", f path)
			d := f open readLines detect(line, line containsSeq("docDependsOn"))
			f close

			if(d) then(
				prerequisitName := "Io" .. d afterSeq("(\"") beforeSeq("\")") .. ".c"
				//writeln(">>> ", f path, " depends on ", prerequisitName)
				prerequisit := orderedFiles detect(of, of name == prerequisitName )
				orderedFiles remove(f)
				orderedFiles insertAfter(f, prerequisit)
			)
		)

		iocFiles = orderedFiles

		iocFiles foreach(f,
			initFile write("IoObject *" .. f name fileName .. "_proto(void *state);\n")
		)

		extraFiles foreach(f,
			initFile write("void " .. f name fileName .. "Init(void *context);\n")
		)

		initFile write("\nvoid Io" .. folder name .. "Init(IoObject *context)\n")
		initFile write("{\n")
		if(iocFiles size > 0 or ioFiles size > 0,
		initFile write("\tIoState *self = ((IoObject *)context)->tag->state;\n\n")
		)

		iocFiles foreach(f,
			initFile write("\tIoObject_setSlot_to_(context, SIOSYMBOL(\"" .. f name fileName asMutable removePrefix("Io") .. "\"), " .. f name fileName .. "_proto(self));\n\n")
		)

		extraFiles foreach(f,
			initFile write("\t" .. f name fileName .. "Init(context);\n")
		)

		if(ioCodeFolder and isStatic,
			//processIoCodeFor(folder path)
			//processedIoFiles := folder folderNamed("ioprocessed") filesWithExtension("io")
			//processedIoFiles sortInPlace(name) foreach(f, initFile write(codeForIoFile(f)))
			ioFiles foreach(f, initFile write(codeForIoFile(f)))
		)

		initFile write("}\n")
		initFile close
	)

	codeForIoFile := method(f,
		code := Sequence clone
		if (f size > 0,
			code appendSeq("\t{\n\t\tchar *s = ")
			code appendSeq(f contents splitNoEmpties("\n") map(line, "\"" .. line escape .. "\\n\"") join("\n\t\t"))
			code appendSeq(";\n\t\tIoState_on_doCString_withLabel_(self, context, s, \"" .. f name .. "\");\n")
			code appendSeq("\t}\n\n")
		)
		code
	)

	generateDocs := method(
		systemCall("../../_build/binaries/io ../../tools/io/DocsExtractor.io .")
	)

	cleanDocs := method(
		systemCall("rm -f io/zzz_docs.io")
	)
)
