Addon := Object clone do(
    newSlot("rootPath")
    newSlot("name")

    platform := method(System platform asLowercase)
    dllSuffix := method(
      if(platform == "darwin", return "dylib")
      if(platform containsSeq("windows"), return "dll")
      "so"
    )
    dllName := method("libIo" .. name .. "." .. dllSuffix)
    addonPath := method(Path with(rootPath, name))
    dllPath := method(Path with(addonPath, "_build/dll", dllName))
    
    ioFiles := method(
        d := Directory with(addonPath) folderNamed("io")
        files := d files select(path endsWithSeq(".io"))
        files map(name) sort map(name, d fileNamed(name))
    )

    dependencies := method(
        File with(Path with(addonPath, "depends")) contents split(" ")
    )

    loadDependencies := method(
        //writeln(name, " depends on ", dependencies)
        dependencies foreach(d,
            if(Lobby getSlot(d) == nil, 
                //writeln("loading dependency ", d)
                AddonLoader loadAddonNamed(d)
            )
        )
    )

    load := method(
        //writeln("Addon ", name, " loading from ", addonPath)
        loadDependencies
        //writeln(dllPath)
        DynLib clone setPath(dllPath) open call("Io" .. name .. "Init", IoVM)
        //ioFiles foreach(f, writeln("loading ", f path))
        ioFiles foreach(file, IoVM doFile(file path))
        Lobby getSlot(name)
    )

    exists := method(Directory with(addonPath) exists)
    
    addonProtos := method(
        f := File with(Path with(addonPath, "protos")) 
        if(f exists, f contents split, list())
    )
)

AddonLoader := Object clone do(
    debug := false

    searchPaths := list(".", System installPrefix .. "/lib/io/addons")
    appendSearchPath := method(p, searchPaths appendIfAbsent(p); self)

    addons := method(
        searchFolders := searchPaths map(path, Directory with(path))
        addonFolders := searchFolders map(folders) flatten select(fileNames contains("build.io"))
        addons := addonFolders map(f, Addon clone setRootPath(f path pathComponent) setName(f path lastPathComponent))
        addons
    )
    
    addonFor := method(name, 
        r := addons detect(name == name)
        if(r, return r)
        addons detect(addonProtos contains(name))
    )
    
    hasAddonNamed := method(name, addonFor(name) != nil)

    loadAddonNamed := method(name,
        addon := addonFor(name)
        if(addon, addon load, nil)
        Lobby getSlot(name)
    )
)
