Lobby IoVM do(
	ImmutableSequence := ""
	String := ImmutableSequence
)

Sequence do(

	setSlot("..", method(arg, self asString cloneAppendSeq(arg asString)))

    docSlot("repeated(n)", "Returns a new sequence containing the receiver repeated n number of times.")
    repeated := method(n,
        s := Sequence clone
        n repeatTimes(s appendSeq(self))
        s 
    )

    docSlot("alignLeftInPlace(width, [padding])", "Same as align left but operation is performed on the receiver.")
    alignLeftInPlace := method(width, padding,
        originalSize := size
        padding = padding ifNilEval(" ")
        ((width - size) / padding size) ceil repeatTimes(appendSeq(padding))
        setSize(width max(originalSize))
    )
    
    docSlot("removeSeq(aSequence)", "Removes occurances of aSequence from the receiver.")
    removeSeq := method(s, self replaceSeq(s, ""))

    docSlot("alignLeft(width, [padding])", """
    Example:
    <pre>
    Io> "abc" alignLeft(10, "-")
    ==> abc-------
    Io> "abc" alignLeft(10, "-=")
    ==> abc-=-=-=-
    </pre>
    """)
    alignLeft := method(width, padding,
        asMutable alignLeftInPlace(width, padding)
    )

    docSlot("alignRight(width, [padding])", """
    Example:
    <pre>
    Io> "abc" alignRight(10, "-")
    ==> -------abc
    Io> "abc" alignRight(10, "-=")
    ==> -=-=-=-abc
    </pre>
    """)
    alignRight := method(width, padding,
        Sequence clone alignLeftInPlace(width - size, padding) appendSeq(self)
    )

    docSlot("alignCenter(width, [padding])", """
    Example:
    <pre>
    Io> "abc" alignCenter(10, "-")
    ==> ---abc----
    Io> "abc" alignCenter(10, "-=")
    ==> -=-abc-=-=
    </pre>
    """)
    alignCenter := method(width, padding,
        alignRight(((size + width) / 2) floor, padding) alignLeftInPlace(width, padding)
    )
    
    asSimpleString := method("\"" .. self asString .. "\"")

	splitNoEmpties := method(
		docSlot("split(optionalArg1, optionalArg2, ...)", """
		Returns a list containing the non-empty sub-sequences of the receiver divided by the given arguments.
		If no arguments are given the sequence is split on white space.
		Examples:
		<pre>
		"a   b  c d" splitNoEmpties => list("a", "b", "c", "d")
		"a***b**c*d" splitNoEmpties("*") => list("a", "b", "c", "d")
		"a***b||c,d" splitNoEmpties("*", "|", ",") => list("a", "b", "c", "d")
		</pre>
		""")
	
		self performWithArgList("split", call evalArgs) selectInPlace(size != 0)
	)

  docSlot("findNthSeq(aSequence, n)", "Returns a number with the nth occurence of aSequence")
  findNthSeq := method(str, n,
    num := self findSeq(str)
    if(num isNil, return nil)
    if(n == 1, return num)
    num + self slice(num + 1, self size) findNthSeq(str, n - 1)
  ) 

	docSlot("interpolate(ctx)", "Returns immutable copy of self with interpolateInPlace(ctx) passed to the copy.")
	interpolate := method(context,
		if(context isNil, context = call sender)
		self asMutable interpolateInPlace(context) asSymbol
	)
	
        orderedSplit := method(
            separators := call evalArgs
            if(separators size == 0, return list(self))
            i := 0
            skipped := 0
            r := list
            separators foreach(separator,
                j := findSeq(separator, i) ifNil(
                    skipped = skipped + 1
                    continue
                )
                r append(slice(i, j))
                if(skipped > 0,
                    skipped repeatTimes(r append(nil))
                    skipped = 0
                )
                i = j + separator size
            )
            if(size == 0,
                r append(nil)
            ,
                r append(slice(i))
            )
            skipped repeatTimes(r append(nil))
            r
        )

	docSlot("prependSeq(aSequence1, aSequence2, ...)", "Prepends given sequences in reverse order to the receiver.  Returns self.")
	prependSeq := method(
		(call argCount < 1) ifTrue(return self)
		toPrepend := "" asMutable
		0 to(call argCount - 1) foreach(argIndex,
			arg := call evalArgAt(argIndex)
			if(arg isKindOf(self) or arg isKindOf(ImmutableSequence),
				toPrepend appendSeq(arg)
			,
				Exception raise("Arguments to Sequence prependSeq must be of type Sequence.")
			)
		)
		if(size + toPrepend size == 0, return self)
		toPrepend size repeatTimes(self appendSeq(" "))
		0 to(size - toPrepend size - 1) asList reverseForeach(v,
			atPut(v + toPrepend size, at(v))
		)
		toPrepend foreach(i, v, atPut(i, v))
		self
	)

	sequenceSets := Map clone do(
		atPut("lowercaseSequence",
			lst := list
			"abcdefghijklmnopqrstuvwxyz" foreach(v, lst append(v))
			lst
		)
		atPut("uppercaseSequence",
			lst := list
			at("lowercaseSequence") foreach(v, lst append(v asUppercase))
			lst
		)
		atPut("digitSequence",
			lst := list
			"0123456789" foreach(v, lst append(v))
			lst
		)
		removeSlot("lst")
		removeSlot("v")
	)
	
	docSlot("nextInSequence(skipVal)", "Returns the next item in the sequence.  The optional skipVal parameter allows you to skip ahead skipVal places.")
	nextInSequence := method(skipVal,
		str := self clone asMutable
		skipVal ifNil(skipVal = 1)
		
		(size < 1) or (skipVal < 1) ifTrue(return str asSymbol)
		
		leadingNonNextableChars := 0
		str foreach(char,
			done := false
			countIt := true
			sequenceSets foreach(set,
				set indexOf(char) ifNonNil(countIt = false; done = true)
			)
			if(countIt, leadingNonNextableChars = leadingNonNextableChars + 1)
			if(done, break)
		)
		(leadingNonNextableChars == str size) ifTrue(return str asSymbol)
		
		str := str splitAt(leadingNonNextableChars)
		leadingNonNextables := str at(0)
		str := str at(1) asMutable
		
		0 to(str size - 1) asList reverseForeach(index,
			done := false
			sequenceSets foreach(setName, set,
				x := set indexOf(str at(index))
				if(x,
					x = x + 1
				,
					continue
				)
				if(x == set size,
					str atPut(index, set at(0))
					if(index == 0,
						str prependSeq(set at(if(setName == "digitSequence", 1, 0)) asCharacter)
						done = true
					)
				,
					str atPut(index, set at(x))
					done = true
				)
				break
			)
			if(done, break)
		)
		
		str = leadingNonNextables .. str
		skipVal = skipVal - 1
		if(skipVal < 1,
			str asSymbol
		,
			str nextInSequence(skipVal)
		)
	)
/*	
	docSlot("reverse", "Returns a reversed copy of the receiver.")
	reverse := method(
		asMutable reverseInPlace
	)

    Moved to C	
	docSlot("reverseInPlace", "Reverses the reveiver.  Returns self.")
	reverseInPlace := method(
		if(size < 2, return self)
		s := Sequence clone
		0 to(size - 1) asList reverseForeach(index,
			s appendSeq(self at(index) asCharacter)
		)
		self = s
	)
*/
	docSlot("to(endpoint)", "Convenience constructor that returns a range of sequences from the receiver to the endpoint argument. Increments over each item in that range by 1.")
	to := method(e, toBy(e, 1))   

	docSlot("toBy(endpoint, increment)", "Convenience constructor that returns a range of sequences from the receiver to the endpoint argument. Increments over each item in that range by the value of the increment parameter. The increment parameter must be positive.")
	toBy := method(e, i,
		toByCompare := method(a, b,
			(a size == b size) ifTrue(return a compare(b))
			nulls := "" asMutable
			(a size - b size) abs repeatTimes(nulls appendSeq("\0"))
			if(a size < b size,
				a = nulls .. a
			,
				b = nulls .. b
			)
			a compare(b)
		)
		removeLeadingZerosAfterLeadingSymbols := method(str,
			if(str isMutable not, str = str asMutable)
			
			# skip over leading symbols
			from := 0
			str foreach(char,
				if(char isInASequenceSet, break, from = from + 1)
			)
			# slice off leading '0's
			to := from - 1
			str slice(from) foreach(char, if(char == "0" at(0), to = to + 1, break))
			if(from <= to, str removeSlice(from, to), str)
		)
		

		from := removeLeadingZerosAfterLeadingSymbols(self asMutable strip)
		to := removeLeadingZerosAfterLeadingSymbols(e asMutable strip)

		#if(i < 0, Exception raise("increment value must be non-negative"))
		(i == 1) ifFalse(Exception raise("increment value must be 1"))
		if(toByCompare(from, to) > 0, Exception raise("range must increase"))

		symbols := list
		1 to(from size min(to size)) foreach(index,
			index = 0 - index
			(from at(index) isInASequenceSet == to at(index) isInASequenceSet) ifFalse(
				Exception raise("invalid range")
			)
		)

		if(from size != to size,
			smaller := bigger := nil
			if(from size > to size,
				bigger = from; smaller = to
			,
				bigger = to; smaller = from
			)
			growthChars := bigger slice(0, bigger size - (smaller size))
			setOfFirstCharInSmaller := nil
			sequenceSets foreach(set,
				if(smaller at(0) in(set), setOfFirstCharInSmaller := set; break)
			)
			growthChars foreach(v,
				v in(setOfFirstCharInSmaller) ifFalse(
					Exception raise("invalid range")
				)
			)
		)

		Range clone setRange(from, to, i)
	)

	docSlot("levenshtein(other)", "Returns the levenshtein distance to other.")
	levenshtein := method(other,
		if(other size < self size, return other levenshtein(self))
		current := 0 to(self size) asList
		other foreach(i, otheri,
			previous := current
			current = List clone with(i + 1)
			self foreach(j, selfj,
				current append((current at(j) + 1) min(previous at(j + 1) + 1) min(previous at(j) + if(selfj == otheri, 0, 1)))
			)
		)
		current last
	)
)
