File do(
	docSlot("with(aPath)", "Returns a new instance with the provided path.")
	with := method(path, self clone setPath(path))

	newSlot("streamDestination")
	newSlot("streamReadSize", 1024*64)
	startStreaming := method(streamTo(streamDestination))

	streamTo := method(streamDestination,
		b := Sequence clone
		self open
		while(isAtEnd not, 
			b empty
			readToBufferLength(b, streamReadSize)
			streamDestination write(b)
			yield
		)
	)

	asCSV := method(
		//file        = record *(CRLF record) [CRLF]
		//record      = field *(COMMA field)
		//field       = (escaped / non_escaped)
		//escaped     = DQUOTE *(TEXTDATA / COMMA / CR / LF / 2DQUOTE) DQUOTE
		//non_escaped = *TEXTDATA
		//COMMA       = %x2C
		//CR          = %x0D ;as per section 6.1 of RFC 2234 [2]
		//DQUOTE      = %x22 ;as per section 6.1 of RFC 2234 [2]
		//LF          = %x0A ;as per section 6.1 of RFC 2234 [2]
		//CRLF        = CR LF ;as per section 6.1 of RFC 2234 [2]
		//TEXTDATA    = %x20-21 / %x23-2B / %x2D-7E

		records := List clone
		index := 0

		file := method(
			record
			while(CRLF and self at(index), record)
		)

		record := method(
			records append(List clone)
			field
			while(COMMA, field)
		)

		field := method(
			records last append(Sequence clone)
			if(DQUOTE, escaped, non_escaped)
		)

		escaped := method(
			while(DQUOTE not, TEXTDATA or COMMA or CR or LF or 2DQUOTE ; records last last append(self at(index - 1)))
		)

		non_escaped := method(
			while(TEXTDATA, records last last append(self at(index - 1)))
		)

		COMMA := method(
			if(self at(index) == 0x2C, index = index + 1 ; true, false)
		)

		CR := method(
			if(self at(index) == 0x0D, index = index + 1 ; true, false)
		)

		DQUOTE := method(
			if(self at(index) == 0x22, index = index + 1 ; true, false)
		)

		2DQUOTE := method(
			save := index
			if(DQUOTE and DQUOTE, true, index = save ; false)
		)

		LF := method(
			if(self at(index) == 0x0A, index = index + 1 ; true, false)
		)

		CRLF := method(
			save := index
			if(CR and LF, true, index = save ; false)
		)

		TEXTDATA := method(
			if(0x20 <= self at(index) and self at(index) <= 0x21, index = index + 1 ; return true)
			if(0x23 <= self at(index) and self at(index) <= 0x2B, index = index + 1 ; return true)
			if(0x2D <= self at(index) and self at(index) <= 0x7E, index = index + 1 ; return true)
			false
		)

		file
		records
	)
)


